#include <exec/types.h>
#include <exec/memory.h>
#include <exec/nodes.h>
#include <libraries/dos.h>
#include <libraries/dosextens.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h> /* added for tolower() prototype -- EDB */
#ifndef __GNUC__
#include <clib/exec_protos.h>
#include <clib/alib_protos.h>
#include <clib/dos_protos.h>
#endif
#ifdef __GNUC__
#include <proto/exec.h>
#include <proto/dos.h>
#endif
#ifdef __SASC
#include <proto/exec.h>
#include <proto/dos.h>
#endif

#include "DirScanner.h"
#include "AmigaMem.h"

#define PATHNAME_SIZE 128
   /* 128 characters is hopefully big enough for any file path. */

#define ALLOCNODE (struct Node*) AllocMem( sizeof( struct Node ), MEMF_CLEAR | MEMF_PUBLIC)
#define FREENODE(n)  FreeMem( n, sizeof(struct Node) )

#define DupString(s) Astrdup(s)
#define FreeString(s) Afree(s)

void DirScanner_DeviceScan( struct DirScanner *ds );


char *DS_FullPathName( struct FileLock      *CurrentDirLock,
                       struct FileInfoBlock *fib,
                       char                 *buffer )
{
   struct FileLock *parentlock;

   if( parentlock = (struct FileLock *)ParentDir( (BPTR) CurrentDirLock ) )
   {
      DS_FullPathName( parentlock, fib, buffer );
      UnLock((BPTR) parentlock);

      Examine( (BPTR) CurrentDirLock, fib );

      strcat( buffer, fib->fib_FileName );
      if( fib->fib_DirEntryType > 0 )
         strcat( buffer, "/" );

   }
   else
   {
      /* No parent, this must be the root dir on the volume.  */

      Examine( (BPTR) CurrentDirLock, fib );
      strcpy( buffer, fib->fib_FileName );
      strcat( buffer, ":" );

   }

   return buffer;
}


char* DS_PathFromLock( struct FileLock *lock )
{
   struct FileInfoBlock *fib;
   char *buffer = NULL;

   if( ( fib = (struct FileInfoBlock*)
         AllocMem( sizeof( struct FileInfoBlock ), MEMF_CLEAR | MEMF_PUBLIC ) )
      != NULL )
   {
      if( buffer = Amalloc( sizeof( 129 ) ) )
      {
         if( DS_FullPathName( lock, fib, buffer ) == NULL )
         {
            Afree( buffer );
            buffer=NULL;
         }
      }

      FreeMem( fib, sizeof( struct FileInfoBlock ) );
   }
   return buffer;
}


BOOL DirScanner_Init( struct DirScanner *ds,
                      DS_Filter         filter )
{
   ds->mode     = DS_MODE_READING;
   ds->pathlock = NULL;
   ds->filter   = filter;

   NewList( (struct List *) &ds->list );

   ds->fib = (struct FileInfoBlock*) AllocMem( sizeof( struct FileInfoBlock ),
                  MEMF_CLEAR | MEMF_PUBLIC );

   if( ds->fib == NULL ) return FALSE;

   ds->pathname = (char*) AllocMem( PATHNAME_SIZE, MEMF_CLEAR | MEMF_PUBLIC );

   if( ds->pathname == NULL ) return FALSE;

   ds->pathname[0] = '\0';

   DirScanner_SetPath( ds, NULL );

   return (BOOL) TRUE;
}


void DirScanner_FreeList( struct DirScanner *ds )
{
   struct Node *node, *next_node;

   for (node = (struct Node*) ds->list.mlh_Head;
        node->ln_Succ != NULL;
        node = next_node )
   {
      next_node = node->ln_Succ;

      FreeString(node->ln_Name);
      FREENODE(node);
   }

   NewList( (struct List*)&ds->list );
}


void DirScanner_AbandonPath( struct DirScanner *ds )
{
   DirScanner_FreeList( ds );

   if( ds->pathlock )
   {
      UnLock( (BPTR) ds->pathlock );
      ds->pathlock = NULL;
      ds->pathname[0] = '\0';
   }
   ds->mode = DS_MODE_WAITING;
}


BOOL DirScanner_SetParentPath( struct DirScanner *ds )
{
   struct FileLock *parentlock;

   if( ds->pathlock )
   {

      if( parentlock = (struct FileLock *)ParentDir( (BPTR) ds->pathlock ) )
      {
         DirScanner_AbandonPath( ds );
         ds->pathlock = parentlock;
         ds->mode     = DS_MODE_READING;
         Examine( (BPTR) parentlock, ds->fib );

         DS_FullPathName( parentlock,
                          ds->fib,
                          ds->pathname );
         return TRUE;
      }
      else
      {
#if 0
         DirScanner_AbandonPath( ds );
         DirScanner_DeviceScan( ds );
         return TRUE;
#endif
         return FALSE;
      }

   }

   return FALSE;
}

BOOL DirScanner_SetChildPath( struct DirScanner *ds,
                              struct Node       *node )
{
   short last;

   if( node->ln_Type )
   {
      char *subdir_name;

      if( subdir_name = DupString( node->ln_Name ) )
      {
         /* 'node->ln_Name' becomes garbage as soon as _AbandonPath()
            is called.  We need a temp copy.
         */
     /* EDB -- This caused any attempts to enter a directory beyond the */
     /* first sucessful one to fail due to not having a path so I commented */
     /* it out.  "Parent" worked fine but then if user selected a directory */
     /* no go, or if the user tried to go two directory levels deep no go */

     /*    DirScanner_AbandonPath( ds );   */

         /* Make sure there is a '/' separator between dirs. */
         last = strlen( ds->pathname ) -1;

         if( ( last >= 0 )     /* pathname is not empty. */
            && ( ds->pathname[last] != '/' ) && ( ds->pathname[last] != ':' ) )
            strcat( ds->pathname, "/" );

         strcat( ds->pathname, subdir_name );
         FreeString(subdir_name);

         ds->pathlock = (struct FileLock *)Lock( ds->pathname, ACCESS_READ );

         /*
            Even though 'node' is allegedly a directory from the list
            we still need to perform the following checks, b/c the user
            might have deleted that directory between the last scan and
            this operation.
         */

         if( ds->pathlock ) /* pathname exists. */
         {
            Examine( (BPTR) ds->pathlock, ds->fib );

            if( ds->fib->fib_DirEntryType > 0 ) /* its a directory. */
            {
               ds->mode = DS_MODE_READING;
               return TRUE;
            }
            else /* this file is not a directory. */
            {
               UnLock( (BPTR) ds->pathlock );
               ds->mode = DS_MODE_WAITING;
            }
         }
         else /* path does not exist. */
         {
            ds->mode = DS_MODE_WAITING;
         }

      } /* End if( subdir_name = DupString( node->ln_Name ) ) */

   } /* End if( node->ln_Type ) */
   return FALSE;
}



BOOL DirScanner_SetPath( struct DirScanner *ds, char *path )
{

   if( DS_PARENTPATH == path )
      return DirScanner_SetParentPath( ds );

   if( DS_DEVICEPATH == path )
   {
      DirScanner_DeviceScan( ds );
      return TRUE;
   }

   if( NULL == path )
       path = "";

   DirScanner_AbandonPath( ds );

   ds->pathlock = (struct FileLock *)Lock( path, ACCESS_READ );

   if( ds->pathlock )
   {
      DS_FullPathName( ds->pathlock, ds->fib, ds->pathname );

      Examine( (BPTR) ds->pathlock, ds->fib );

      if (ds->fib->fib_DirEntryType > 0)
      {
         ds->mode = DS_MODE_READING;
         return TRUE;
      }
      else /* this file is not a directory. */
      {
         UnLock( (BPTR) ds->pathlock );
         ds->mode = DS_MODE_WAITING;
         return FALSE;
      }
   }
   else /* path does not exists. */
   {
      ds->mode = DS_MODE_WAITING;
      return FALSE;
   }
}


struct Node*
DirScanner_Scan( struct DirScanner *ds )
{
   struct Node *result = NULL;

   if( ds->mode )
   {
      if( ExNext( (BPTR) ds->pathlock, ds->fib ) )
      {
         if( ( ds->filter == NULL ) || ( (*ds->filter)(ds->fib) ) ) /* Passes filter. */
         {
            struct Node *node;

            if( node = ALLOCNODE )
            {
               if( node->ln_Name = DupString( ds->fib->fib_FileName ) )
               {
                  node->ln_Type = ( ds->fib->fib_DirEntryType > 0 );

                  AddTail( (struct List*) &ds->list, node );

                  result = node;
               }
               else /* couldn't allocate filename string. */
               {
                  FREENODE( node );
                  ds->mode = DS_MODE_WAITING;
               }
            }
         }

      }
      else /* no more files in the directory. */
         ds->mode = DS_MODE_WAITING;
   }

   return result;
}


void DirScanner_CleanUp( struct DirScanner *ds )
{
   DirScanner_AbandonPath( ds );

   if( ds->fib )
      FreeMem( ds->fib, sizeof( struct FileInfoBlock ) );

   if( ds->pathname )
      FreeMem( ds->pathname, PATHNAME_SIZE );

}



char *DirScanner_BSTRdup( BSTR bstr )
{
   char *astr;
   char *dup;
   LONG length;

   astr = (char *)( bstr << 2 );

   length = astr[0];

   if( dup = (char*) Amalloc( length +2 ) )
   {
      strncpy( dup, (char*) &astr[1], length );
      dup[length]   = ':';
      dup[length+1] = '\0';
   }

   return dup;
}

void DirScanner_DeviceScan( struct DirScanner *ds )
{
   struct DosLibrary *DosBase;
   struct RootNode *root;
   struct DosInfo *info;
   struct DosList *list;

   DirScanner_AbandonPath( ds );

   ds->pathname[0] = '\0'; /* terminate the pathname. */

   DosBase = (struct DosLibrary*) OpenLibrary(DOSNAME,0);
   root    = (void *)(DosBase->dl_Root);
   info    = (void *)(root->rn_Info << 2);
   list    = (void *)(info->di_DevInfo << 2);

   while( list != NULL )
   {
      struct Node *node;

      if( node = ALLOCNODE )
      {
         if( node->ln_Name = DirScanner_BSTRdup( list->dol_Name ) )
         {
            node->ln_Type = 2;
            AddTail( (struct List*) &ds->list, node );
         }
         else /* couldn't allocate filename string. */
         {
            FREENODE(node);
            ds->mode = DS_MODE_WAITING;
         }
      }

      list = (void *)(list->dol_Next << 2);
   }

   CloseLibrary((struct Library*) DosBase);

   ds->mode = DS_MODE_WAITING;

}



BOOL DSFilter_NoIcons( struct FileInfoBlock *fib )
{
   short length, i;
   char *filename;
   char suffix[6];

   filename = fib->fib_FileName;
   length = strlen( filename );
   /* EDB -- added '=' so files named ".info" will be ignored also */
   if( length >= 5 )
   {
      strcpy( suffix, &filename[length-5] );
      for( i = 0; i < 5; i++ )     /* convert 'suffix' to all lower case. */
         suffix[i] = tolower( suffix[i] );

      return (BOOL) strcmp( suffix, ".info" );
   }

   return TRUE;
}


