
#include "pcg_FileReq.h"

#include <exec/memory.h>
#include <exec/nodes.h>
#include <stdio.h>
#include <string.h>
#ifndef __GNUC__
#include <clib/exec_protos.h>
#include <clib/intuition_protos.h>
#include <clib/graphics_protos.h>
#endif
#ifdef __GNUC__
#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#endif
#ifdef __SASC
#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#endif

#include "amigamem.h"
#include "dirscanner.h"

#include "dumpwait.h"

struct Lock* ParentDir( struct Lock * );
struct Lock* Lock( char*, int );


#define QUALFLAG (BYTE*) -1

#define STRDUP(s) Astrdup(s)

#define FREE(m) if( m ) Afree( m )


typedef struct FileReqWindow
{
   pcgWindow     window;
   ScrollingList scrollinglist;
   StringGadget  sg_drawer;
   StringGadget  sg_file;
   BoolGadget    bg_ok;
   BoolGadget    bg_disks;
   BoolGadget    bg_parent;
   BoolGadget    bg_cancel;

} FileReqWindow;

#define ANYSIZE 65535

void FileReqWindow_Init( FileReqWindow *frw,
                         Screen        *screen,
                         char          *title,
                         MsgPort       *SharedUserPort )
{
   pcg_3DPens pens;

/*   pens = StandardPens();  */
/* Prepare for user defined screens -- EDB */

   pens = StandardScreenPens( screen );

   /* Initialize the frw.window */ /* changed to sizebbottom -- EDB */
   pcgWindow_Init( &frw->window,
      0, 12, 320, 200, /* LeftEdge, TopEdge, Width, Height */
      280, 150, ANYSIZE, ANYSIZE, /* min width/height, max width/height */
      title,
      REFRESHWINDOW | CLOSEWINDOW | NEWSIZE, /* IDCMPFlags */
      WINDOWSIZING | SIZEBBOTTOM | WINDOWDRAG | WINDOWDEPTH |
      WINDOWCLOSE | SIMPLE_REFRESH | ACTIVATE, /* Flags */
      screen ); /* pointer to screen */
   frw->window.SharedUserPort      = SharedUserPort;

   ScrollingList_Init( &frw->scrollinglist,
      8, 16, 246, 120, /* LeftEdge, TopEdge, Width, Height */
      pens, FALSE );    /* Allow Selection of >1 item? */
   AddInteractor( &frw->window, (Interactor*) &frw->scrollinglist );

   StringGadget_Init( &frw->sg_drawer,
      72, 144, 184, /* LeftEdge, TopEdge, Width */
      80, /* # of characters */
      pens, "Drawer:" );
   AddInteractor( &frw->window, (Interactor*) &frw->sg_drawer );

   StringGadget_Init( &frw->sg_file,
      72, 160, 184, /* LeftEdge, TopEdge, Width */
      36, /* # of characters */
      pens, "File:" );
   AddInteractor( &frw->window, (Interactor*) &frw->sg_file );

   BoolGadget_Init( &frw->bg_ok,
      8, 176, 56, 14, /* LeftEdge, TopEdge, Width, Height */
      pens, "OK" );
   AddInteractor( &frw->window, (Interactor*) &frw->bg_ok );

   BoolGadget_Init( &frw->bg_disks,
      72, 176, 56, 14, /* LeftEdge, TopEdge, Width, Height */
      pens, "Disks" );
   AddInteractor( &frw->window, (Interactor*) &frw->bg_disks );

   BoolGadget_Init( &frw->bg_parent,
      136, 176, 56, 14, /* LeftEdge, TopEdge, Width, Height */
      pens, "Parent" );
   AddInteractor( &frw->window, (Interactor*) &frw->bg_parent );

   BoolGadget_Init( &frw->bg_cancel,
      200, 176, 56, 14, /* LeftEdge, TopEdge, Width, Height */
      pens, "Cancel" );
   AddInteractor( &frw->window, (Interactor*) &frw->bg_cancel );
}


#define BORDERPAD 4
#define BUTTON_POS(i) ( leftborder + ( ( useable_size.x - size.x ) * i )/3 )

void FileReqWindow_Resize( FileReqWindow *frw )
{
   Point window_size, size, loc;
   Point useable_size;
   Window *iwindow;
   short leftborder, rightborder, topborder, bottomborder;
   short y;


   iwindow      = iWindow( &frw->window );
   leftborder   = iwindow->BorderLeft   + BORDERPAD;
   rightborder  = iwindow->BorderRight  + BORDERPAD;
   topborder    = iwindow->BorderTop    + BORDERPAD;
   bottomborder = iwindow->BorderBottom + BORDERPAD;

   window_size    = Size( (struct GraphicObject *)&frw->window );
   useable_size.x = window_size.x - leftborder - rightborder;
   useable_size.y = window_size.y - topborder - bottomborder;
      /* window_size is now the useable size of the window. */

   size  = Size( (struct GraphicObject *)&frw->bg_ok );

   y = window_size.y - bottomborder - size.y;

   SetLocation( (struct GraphicObject *)&frw->bg_ok,     BUTTON_POS(0), y );
   SetLocation( (struct GraphicObject *)&frw->bg_disks,  BUTTON_POS(1), y );
   SetLocation( (struct GraphicObject *)&frw->bg_parent, BUTTON_POS(2), y );
   SetLocation( (struct GraphicObject *)&frw->bg_cancel, BUTTON_POS(3), y );

   size = Size( (struct GraphicObject *)&frw->sg_file );
   loc  = Location( (struct GraphicObject *)&frw->sg_file );
   y -= size.y + BORDERPAD;
   SetLocation( (struct GraphicObject *)&frw->sg_file, loc.x, y );
   SetSize( (struct GraphicObject *)&frw->sg_file, useable_size.x - loc.x + BORDERPAD, size.y );

   size = Size( (struct GraphicObject *)&frw->sg_drawer );
   loc  = Location( (struct GraphicObject *)&frw->sg_drawer );
   y -= size.y + BORDERPAD;
   SetLocation( (struct GraphicObject *)&frw->sg_drawer, loc.x, y );
   SetSize( (struct GraphicObject *)&frw->sg_drawer, useable_size.x - loc.x + BORDERPAD, size.y );


   loc.x = leftborder;
   loc.y = topborder;
   y -= BORDERPAD;
   SetLocation( (struct GraphicObject *)&frw->scrollinglist, loc.x, loc.y );
   SetSize( (struct GraphicObject *)&frw->scrollinglist, useable_size.x, y - loc.y );

   SetDrMd( iwindow->RPort, JAM1 );
   SetAPen( iwindow->RPort, 0 );

   RectFill( iwindow->RPort, 0,0, window_size.x, window_size.y );
   RefreshWindowFrame( iwindow );

   Refresh( (struct Interactor *)&frw->window );
}


void
pcgfr_SetPath( struct FileReqWindow *frw,
               struct DirScanner *scanner,
               char *path )
{
   int len = strlen( path );
   int i;

   if( len == 0 ) return;

   for( i = len-1; i >= 0; i-- )
   {
      if( path[i] == '/' || path[i] == ':' )
          break;
   }

   if( i == len -1 ) /* path is all path */
   {
      SetStringValue( (struct Interactor *)&frw->sg_drawer, path );
      DirScanner_SetPath( scanner, path );
   }
   else if( i < 0 ) /* path is a filename only, no path. */
   {
      SetStringValue( (struct Interactor *)&frw->sg_file, path );
   }
   else
   {
      char c;   /*  , *file;  */

      c = path[i+1];
      path[i+1] = '\0';

      SetStringValue( (struct Interactor *)&frw->sg_drawer, path );
      DirScanner_SetPath( scanner, path );

      path[i+1] = c;
      SetStringValue( (struct Interactor *)&frw->sg_file, &path[i+1] );

   }
}


void
pcgfr_GetPath( struct FileReqWindow *frw, struct DirScanner *scanner )
{
   SetStringValue( (struct Interactor *)&frw->sg_drawer, scanner->pathname );
}


BOOL pcg_FileReq( FileReqInfo *frInfo )
{
   FileReqWindow        frw;

   BOOL                  done;
   BOOL                  result;
   IntuiMessage         *message, event;
   StringList           *stringlist;
   struct MsgPort       *port;
   Window               *iwindow;
   static ULONG          glsecs = 0L, glmics = 0L; /* for gadget doubleclick */


   struct DirScanner    scanner;

/* DUMPWAIT("pcg_FileReq : before DirScanner_Init\n"); */

   DirScanner_Init( &scanner, DSFilter_NoIcons );

/* DUMPWAIT("before FileReqWindow_Init\n"); */

   FileReqWindow_Init( &frw, frInfo->screen, frInfo->title,
                        frInfo->SharedUserPort );

/* DUMPWAIT("before pcgfr_SetPath\n"); */

   pcgfr_SetPath( &frw, &scanner, frInfo->FullPathName );

/* DUMPWAIT("before pcgfr_GetPath\n"); */

   pcgfr_GetPath( &frw, &scanner );

   stringlist = StringList_of( (struct Interactor *)&frw.scrollinglist );

/* DUMPWAIT("before opening window.\n"); */

   /*  Open the frw.window (returns NULL on failure) */
   if( pcgOpenWindow( &frw.window ) )
   {
      iwindow = iWindow( &frw.window );
      port    = iwindow->UserPort;
      FileReqWindow_Resize( &frw );


      Activate( (struct Interactor *)&frw.sg_file, TRUE );
      done = FALSE;
      while( !done ) /* Main event loop */
      {
         /* printf("top of message loop\n"); */

         if (scanner.mode == DS_MODE_READING)
         {
            struct Node *node;

            node = DirScanner_Scan( &scanner );

            if( node != NULL ) /* Put this file into the list. */
            {
               AddString( (struct Interactor *)&frw.scrollinglist,
                     node->ln_Name,
                     (UBYTE)( ( node->ln_Type ) ? ENTRY_SPECIAL : 0 )
                  );

               Refresh( (struct Interactor *)&frw.scrollinglist );
            }
         }

         message = (IntuiMessage *)GetMsg( port );

         if( message == NULL )
         {
            if( scanner.mode == DS_MODE_WAITING )
            {
               WaitPort( port );
            }
            continue;
         }


         event   = *message; /* Make a local copy of the message */
         ReplyMsg( (struct Message *)message );

         switch( event.Class )
         {

            case CLOSEWINDOW:
               done   = TRUE; /* exit the loop.  */
               result = FALSE;
               break;

            case REFRESHWINDOW:
               {
                  pcgWindow *some_window;

                  /*           REMEMBER:
                  ** You may be refreshing a window other than
                  ** the file requester itself!
                  */
                  some_window = (pcgWindow *)event.IDCMPWindow->UserData;
                  BeginRefresh( iWindow( some_window ) );
                  Refresh( (struct Interactor *)some_window );
                  EndRefresh( iWindow( some_window ), TRUE );
               }
               break;

            case NEWSIZE:
               /* resize the gadgets to fill the window. */
               FileReqWindow_Resize( &frw );
               break;

            default:
               if( Respond( (Interactor *)&frw.bg_ok, &event )
                   & CHANGED_STATE )
               {  /* frw.bg_ok was hit. */

                  if( Value( (struct Interactor *)&frw.sg_file ) )
                  {
                     result = TRUE;
                     done   = TRUE;
                  }
                  else
                  {
                    DisplayBeep( frInfo->screen );
                     Activate( (struct Interactor *)&frw.sg_file, TRUE );
                  }
               }

               if( Respond( (Interactor *)&frw.bg_cancel, &event )
                   & CHANGED_STATE )
               {  /* frw.bg_cancel was hit. */
                  done   = TRUE;
                  result = FALSE;
               }

               if( Respond( (Interactor *)&frw.bg_parent, &event )
                   & CHANGED_STATE )
               {  /* frw.bg_parent was hit. */

                  if( DirScanner_SetPath( &scanner, DS_PARENTPATH ) )
                  {
                     pcgfr_GetPath( &frw, &scanner );
                     DeleteAllStrings( (struct Interactor *)&frw.scrollinglist );
                     Refresh( (struct Interactor *)&frw.scrollinglist );
                     Refresh( (struct Interactor *)&frw.sg_drawer );
                  }
                  else
                  {
                     DisplayBeep( frInfo->screen );
                  }
               }

               if( Respond( (Interactor *)&frw.scrollinglist, &event )
                   & CHANGED_STATE )
               { /* An item was selected. */
                  short       i;
                  char*       entry;

                  for( i = 0; i < stringlist->nEntries; i++ )
                  {
                     if( stringlist->Qualifiers[i] & ENTRY_SELECTED )
                     {
                        entry = stringlist->Entries[i];
                        /* printf("entry=%s, qualifier=%d\n", entry,
                                   stringlist->Qualifiers[i] );         */
                        if( stringlist->Qualifiers[i] & ENTRY_SPECIAL )
                        {
                           /* Hit a directory */
                           struct Node *node;

                           for( node = (struct Node*) scanner.list.mlh_Head;
                                node->ln_Succ != NULL;
                                node = node->ln_Succ )
                           {
                              if( 0 == strcmp( node->ln_Name, entry ) )
                              {
                                 break;
                              }
                           }


                           if( ( node->ln_Succ != NULL )
                                && ( node->ln_Name != NULL ) )
                           {
                              if( DirScanner_SetChildPath( &scanner, node ) )
                              {
                                 pcgfr_GetPath( &frw, &scanner );
                                 DeleteAllStrings( (struct Interactor *)&frw.scrollinglist );
                                 Refresh( (struct Interactor *)&frw.scrollinglist );
                                 Refresh( (struct Interactor *)&frw.sg_drawer );
                              }
                              else
                              {
                                 DisplayBeep( frInfo->screen );
                              }
                           }
                           else /* should never happen */
                           {
                              DisplayBeep( frInfo->screen );
                           }

                        } /* End if( stringlist->Qualifiers[i]
                                     & ENTRY_SPECIAL ) */
                        else
                        {
                           /* plain vanilla file */
                           SetStringValue( (Valuator *)&frw.sg_file, entry );
                           Refresh( (Interactor *)&frw.sg_file );

                           if( DoubleClick( glsecs, glmics, event.Seconds, event.Micros ) )
                             {
                                /*
                                 * Double Click on file in listview
                                 * Act as if OK button was pressed
                                 * -- EDB
                                 */
                                if( Value( (struct Interactor *)&frw.sg_file ) )
                                {
                                  result = TRUE;
                                  done   = TRUE;
                                }
                                else
                                {
                                  DisplayBeep( frInfo->screen );
                                  Activate( (struct Interactor *)&frw.sg_file, TRUE );
                                }
                             }
                          else
                             { /* set the time for next check */
                                glsecs = event.Seconds;
                                glmics = event.Micros;
                             }

                        break;
                        }
                     }
                  }
               }

               if( Respond( (Interactor *)&frw.sg_drawer, &event )
                    & ( CHANGED_STATE | DEACTIVATED ) )
               {  /* User entered something into sg_drawer  */
                  /* Activate next string or integer gadget. */

                  if( DirScanner_SetPath( &scanner,
                      StringValue( (struct Interactor *)&frw.sg_drawer ) ) )
                  {
                     pcgfr_GetPath( &frw, &scanner );
                     DeleteAllStrings( (struct Interactor *)&frw.scrollinglist );
                     Refresh( (struct Interactor *)&frw.scrollinglist );
                     Refresh( (struct Interactor *)&frw.sg_drawer );

                     ActivateNext( (Interactor *)&frw.sg_drawer );
                  }

                  else
                  {
                     Activate( (Interactor *)&frw.sg_drawer, TRUE );
                  }


               }

               if( Respond( (Interactor *)&frw.sg_file, &event )
                   & ( CHANGED_STATE | DEACTIVATED ) )
               {  /* User entered something into frw.sg_file */

                     result = TRUE;
                     done   = TRUE;

                  /* Activate next string or integer gadget. */
                  ActivateNext( (Interactor *)&frw.sg_file );
               }

               if( Respond( (Interactor *)&frw.bg_disks, &event )
                   & CHANGED_STATE )
               {  /* frw.bg_disks was hit. */

                  if( DirScanner_SetPath( &scanner, DS_DEVICEPATH ) )
                  {
                     struct Node *node;

                     DeleteAllStrings( (struct Interactor *)&frw.scrollinglist );

                     for( node= (struct Node *)scanner.list.mlh_Head;
                          node->ln_Succ != NULL;
                          node = node->ln_Succ )
                     {
                        AddString( (struct Interactor *)&frw.scrollinglist,
                           node->ln_Name, ENTRY_SPECIAL );
                     }

                     SetStringValue( (struct Interactor *)&frw.sg_drawer, NULL );

                     Refresh( (struct Interactor *)&frw.sg_drawer );
                     Refresh( (struct Interactor *)&frw.scrollinglist );

                   }
               }


         } /* end switch */

      } /* end loop */

      pcgCloseWindow( &frw.window );


      if( result )
      {
         char *fullpath = frInfo->FullPathName;
         char *path = StringValue( (struct Interactor *)&frw.sg_drawer );
         char *file = StringValue( (struct Interactor *)&frw.sg_file );
         int pathlen;

         fullpath[0] = '\0';

         if( path != NULL )
         {
            pathlen = strlen( path );
            strcat( fullpath, path );

            if( path[pathlen-1] != '/' && path[pathlen-1] != ':' )
               strcat( fullpath, "/" );
         }
         strcat( fullpath, file );
      }


   } /* end if */

   else /* Window could not be opened. */
   {
      ; /*** YOUR ERROR HANDLING HERE ***/
      /* DisplayBeep( frInfo->screen ); */
   }


   /* Cleanup */

/* DUMPWAIT("cleaning...\n");  */

   CleanUp( (PObject *)&frw.window );
   CleanUp( (PObject *)&frw.scrollinglist );
   CleanUp( (PObject *)&frw.sg_drawer );
   CleanUp( (PObject *)&frw.sg_file );
   CleanUp( (PObject *)&frw.bg_ok );
   CleanUp( (PObject *)&frw.bg_disks );
   CleanUp( (PObject *)&frw.bg_parent );
   CleanUp( (PObject *)&frw.bg_cancel );

/* DUMPWAIT("will call DirScanner_CleanUp\n"); */
   DirScanner_CleanUp( &scanner );

/* DUMPWAIT("did call DirScanner_CleanUp\n");  */
   return result;
}
